(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrbool eqtype ssrnat seq.
Require Import ssrZ ZArith_ext seq_ext ssrnat_ext uniq_tac.

Require Import bipl seplog integral_type.
Module seplog_Z := Seplog ZIT.

Import seplog_Z.
Import assert_m.expr_m.
Import assert_m.

Local Open Scope seplog_expr_scope.
Local Open Scope seplog_cmd_scope.
Local Open Scope seplog_assert_scope.
Local Open Scope heap_scope.

(** A tentative formalization of
   Automatic Numeric Abstractions for Heap-Manipulating Programs,
   S. Magill, M.-H. Tsai, P. Lee, Y.-K. Tsay, POPL 2010. *)

(** syntax of programs *)

Inductive cont : Type :=
| cont_seq : cmd0 -> cont -> cont 
| cont_nondeter : var.v -> cont -> cont
| cont_halt : cont
| cont_abort : cont
| cont_goto : nat -> cont
| cont_branch : list expr_b -> list cont -> cont.

Notation "c ; d" := (cont_seq c d) (at level 81, right associativity) : abst_scope.
Notation "x '<-?;' c" := (cont_nondeter x c) (at level 81, right associativity) : abst_scope.
Notation "'assume' e ;; k" := (cont_branch (e::nil) (k::nil)) (at level 81, right associativity) : abst_scope.
Local Open Scope abst_scope.

Definition prog := list (nat * cont).

(** the predicate for singly-linked lists and some properties *)

Variable next : Z.

Inductive ls : expr -> expr -> expr -> assert :=
| ls_nil : forall n e1 e2 s h, emp s h ->
  [ n ]e_s = 0 -> 
  [ e1 ]e_s = [ e2 ]e_s ->
  ls n e1 e2 s h
| ls_cons : forall n first tail z s h h1 h2, h1 # h2 -> h = h1 \U h2 ->
  (0 < [ n ]e_s) ->
  (first |~>_ ** first \+ cst_e next |~> z) s h1 ->
  ls (n \- cst_e 1) z tail s h2 ->
  ls n first tail s h .

Lemma ls_nil_inv : forall n e1 e2 s h, ls n e1 e2 s h ->
  [ n ]e_s = 0 ->
  h = heap.emp /\ [ e1 ]e_s = [ e2 ]e_s.
Proof.
move=> n e1 e2 s h.
inversion 1 => //.
subst.
move=> X; rewrite X in H2.
by move/Zlt_irrefl : H2.
Qed.

Lemma ls_cons_inv : forall n e1 e2 s h, ls n e1 e2 s h ->
  0 < [ n ]e_s -> exists h1, exists h2, exists z,
    h1 # h2 /\ h = h1 \U h2 /\
    (e1 |~>_ ** e1 \+ cst_e next |~> z) s h1 /\
    ls (n \- cst_e 1) z e2 s h2.
Proof.
move=> n e1 e2 s h.
inversion 1 => //.
- subst.
  rewrite H1.
  by move/Zlt_irrefl.
- subst => _.
  by exists h1, h2, z.
Qed.

Lemma ls_equiv : forall n s h e1 e2, ls n e1 e2 s h ->
  forall n' e1' e2' s', 
    [ n ]e_s = [ n' ]e_s' -> [ e1 ]e_s = [ e1' ]e_s' -> [ e2 ]e_s = [ e2' ]e_s' ->
    ls n' e1' e2' s' h.
Proof.
move=> n s h e1 e2; elim; clear n s h e1 e2.
- move=> n e1 e2 s h; rewrite /emp => Hh; subst h.
  move=> Hn He1e2 n' e1' e2' s' Hnn' He1e1' He2e2'.
  apply ls_nil => //. 
  by rewrite -Hnn' Hn.
  by rewrite -He1e1' -He2e2'.
- move=> n first tail z s h h1 h2 Hdisj Hunion Hn Hh1 Hh2 IH n' e1' e2' s' Hnn' Hfirste1 Htaile2'.
  apply ls_cons with (cst_e [ z ]e_s) h1 h2 => //.
  by rewrite -Hnn'.
  move: Hh1.
  apply monotony.
  move=> h0 Hh0; by Mapsto.
  move=> h0 Hh0; by Mapsto.
  apply IH => //.
  by rewrite /= Hnn'.
Qed.

Lemma ls_app_cons : forall n1 e1 e2 s h1, ls n1 e1 e2 s h1 ->
  forall z h2, h1 # h2 ->
    (e2 |~>_ ** e2 \+ cst_e next |~> z) s h2 ->
    ls (n1 \+ cst_e 1) e1 z s (h1 \U h2).
Proof.
move=> n1 e1 e2 s h1; elim; clear n1 e1 e2 s h1.
- move=> n1 e1 e2 s h1 Hh1 Hn1 He1e2 z h2 H' H.
  rewrite /emp in Hh1;  subst h1.
  apply ls_cons with z h2 heap.emp => //.
  by apply heap.disjhe.
  by rewrite heap.unionhe heap.unioneh.
  by rewrite /= Hn1 /=.
  move: H.
  apply monotony => //.
  move=> h3 Hh3; by Mapsto.
  move=> h3 Hh3; by Mapsto.
  apply ls_nil => //.
  by rewrite /= Hn1 /=.
- move=> n first tail z s h h1 h2 Hdisj Hunion Hn Hh1 Hh2 IH z' h3 Hh2' Hh3.
  apply ls_cons with z h1 (h2 \U h3) => //.
  by map_tac_m.Disj.
  by map_tac_m.Equal.
  rewrite /= /ZIT.t_plus; omega.
  eapply ls_equiv.
  apply IH; by [map_tac_m.Disj | apply Hh3].
  rewrite /= /ZIT.t_plus /ZIT.t_minus /ZIT.t; omega.
  done.
  done.
Qed.

(** the predicate for trees *)

Definition NIL := cst_e 0.

Variable left right : Z.

Inductive tree : expr -> expr -> assert :=
| tree_nil : forall n r s h, emp s h ->
  [ n ]e_s = 0 -> [ r ]e_s = 0 ->
  tree n r s h
| tree_cons : forall n r s h n1 n2 lc rc h0 h1 h2,
  h0 # h1 -> h1 # h2 -> h0 # h2 -> h = h0 \U h1 \U h2 ->
  (0 < [ n ]e_s) ->
  [ n1 ]e_s < [ n ]e_s ->
  [ n2 ]e_s < [ n ]e_s ->
  (r |~>_ ** r \+ cst_e left |~> lc ** r \+ cst_e right |~> rc)%seplog_assert s h0 ->
  tree n1 lc s h1 ->
  tree n2 rc s h2 ->
  tree n r s h.

Lemma tree_equiv : forall n r s h, tree n r s h ->
  forall n' r' s', [ n ]e_s = [ n' ]e_s' -> [ r ]e_s = [ r' ]e_s' ->
    tree n' r' s' h.
move=> n r s h; elim; clear n r s h.
- move=> n r s h; rewrite /emp => Hh; subst h.
  move=> Hn Hr n' r' s' Hn' Hr'.
  apply tree_nil => //.
  * by rewrite -Hn' Hn.
  * by rewrite -Hr'.
- move=> n r s h n1 n2 lc rc h1 h2 h3 Hdisj12 Hdisj23 Hdisj13 Hunion
Hn Hn1 Hn2 Hr Hlc IH1 Hrc IH2 n' r' s' Hn' Hr'.
  apply: (tree_cons _ _ _ _ (cst_e [ n1 ]e_s) (cst_e [ n2 ]e_s) (cst_e
[ lc ]e_s) (cst_e [ rc ]e_s) h1 h2 h3) => //.
  * by rewrite -Hn'.
  * by rewrite -Hn'.
  * by rewrite -Hn'.
  * apply: (monotony _ _ _ _ _ _ _ _ _ Hr).
    + move=> h4 Hh4.
      by Mapsto.
    + move=> h4.
      apply monotony => //.
      - move=> h5 Hh5; by Mapsto.
      - move=> h6 Hh6; by Mapsto.
  * by apply: IH1.
  * by apply: IH2.
Qed.

(** an extended notion of expression with free and bound variables (locally nameless-like?) *)

Inductive expr' : Type :=
  | fvar_e : var.v -> expr'
  | bvar_e : nat -> expr'
  | cst_e' : ZIT.t -> expr'
  | bop_e' : binop_e -> expr' -> expr' -> expr'
(*  | valabs_e' : expr' -> expr'*)
  | uop_e' : unaop_e -> expr' -> expr'.

Notation "e1 '\+' e2" := (bop_e' add_e e1 e2) (at level 61, left associativity) : abst_scope.
Notation "e1 '\-' e2" := (bop_e' sub_e e1 e2) (at level 61, left associativity) : abst_scope.
Notation "e1 '\*' e2" := (bop_e' mul_e e1 e2) (at level 58, left associativity) : abst_scope.
Definition field' x f := fvar_e x \+ cst_e' f.
Notation "x '-.>' f " := (field' x f) (at level 75) : abst_scope.

Fixpoint expr'2expr e' : option expr :=
  match e' with
    | fvar_e x => Some (var_e x)
    | bvar_e _ => None
    | cst_e' c => Some (cst_e c)
    | bop_e' b e1' e2' => 
      match (expr'2expr e1', expr'2expr e2') with
        | (Some e1, Some e2)  => Some (bop_e b e1 e2)
        | _ => None
      end
(*    | valabs_e' e1' => match expr'2expr e1' with
                        | Some e1 => Some (uop_e valabs_e e1)
                        | _ => None
                      end*)
    | uop_e' u e1' => match expr'2expr e1' with
                        | Some e1 => Some (uop_e u e1)
                        | _ => None
                      end
    end.

Fixpoint inj_expr e : expr' :=
  match e with
    | var_e x => fvar_e x
    | cst_e v => cst_e' v
    | bop_e b e1 e2 => bop_e' b (inj_expr e1) (inj_expr e2)
(*    | uop_e valabs_e e1 => valabs_e' (inj_expr e1)*)
    | uop_e u e1 => uop_e' u (inj_expr e1)
  end.

Definition closed e' :=
  match expr'2expr e' with Some _ => true | None => false end.
    
Lemma closed_bop_e' : forall b e1 e2, closed (bop_e' b e1 e2) -> closed e1 /\ closed e2.
Proof.
move=> b e1 e2; rewrite /closed /=.
case H2 : (expr'2expr e2) => t2; by destruct (expr'2expr e1).
Qed.

Lemma closed_uop_e' : forall e u, closed (uop_e' u e) -> closed e.
Proof. move=> e; rewrite /closed /=; by destruct (expr'2expr e). Qed.

Definition eval' e' s := 
  match expr'2expr e' with 
    | Some e => Some (eval e s)
    | _ => None
  end.

Fixpoint expr'_eq (e1 e2 : expr') : bool :=
  match e1 with
    | fvar_e w1 =>
      match e2 with fvar_e w2 => w1 == w2 | _ => false end
    | bvar_e w1 =>
      match e2 with bvar_e w2 => w1 == w2 | _ => false end
    | cst_e' i1 =>
      match e2 with cst_e' i2 => ZIT.t_eqb i1 i2 | _ => false end
    | bop_e' bo e11 e12 => 
      match e2 with 
        | bop_e' bo' e21 e22 =>
          binop_e_eq bo bo' && expr'_eq e11 e21 && expr'_eq e12 e22 
        | _ => false 
      end
(*    | valabs_e' e11 =>
      match e2 with
        valabs_e' e22 => expr'_eq e11 e22 | _ => false
      end*)
    | uop_e' uo e11 =>
      match e2 with
        uop_e' uo' e22 => unaop_e_eq uo uo' && expr'_eq e11 e22 | _ => false
      end
  end.

Fixpoint subst_e' (e patt repl : expr') : expr' :=
  match e with
    | fvar_e w => match expr'_eq e patt with
                 | true => repl
                 | _ => e
               end
    | bvar_e w => match expr'_eq e patt with
                 | true => repl
                 | _ => e
               end
    | cst_e' i => match expr'_eq e patt with
                   | true => repl
                   | _ => e
                 end
    | bop_e' bo e1 e2 => 
      match expr'_eq e patt with
        | true => repl
        | _ => bop_e' bo (subst_e' e1 patt repl) (subst_e' e2 patt repl)
      end
(*    | valabs_e' e' => match expr'_eq e patt with
                       | true => repl
                       | _ => valabs_e' (subst_e' e' patt repl)
                     end*)
    | uop_e' u e' => match expr'_eq e patt with
                       | true => repl
                       | _ => uop_e' u (subst_e' e' patt repl)
                     end
  end.

Lemma subst_e'_free : forall e i repl,
  closed e -> subst_e' e (bvar_e i) repl = e.
Proof.
elim=> //.
move=> b e1 IH1 e2 IH2 i repl H /=.
case/closed_bop_e' : H => H1 H2.
rewrite IH1 //.
rewrite IH2 //.
move=> uo e IH i repl.
move/closed_uop_e' => H.
by rewrite /= IH.
Qed.

Fixpoint expr'_fvar_set (e : expr') : list var.v :=
  match e with
    | fvar_e x => x :: nil
    | bvar_e _ => nil
    | cst_e' z => nil
    | bop_e' _ e1 e2 =>
      seq_ext.app_set (fun a b => a == b) (expr'_fvar_set e1) (expr'_fvar_set e2)
(*    | valabs_e' e' => expr'_fvar_set e'*)
    | uop_e' _ e' => expr'_fvar_set e'
  end.

Inductive expr_b' : Type :=
| true_b'
| bop_b' : binop_b -> expr' -> expr' -> expr_b'
| neg_b' : expr_b' -> expr_b'
| bop_b2' : binop_b2 -> expr_b' -> expr_b' -> expr_b'.

Notation "e =e e'" := (bop_b' eq_b e e') (at level 76) : abst_scope.
Notation "e <>e e'" := (bop_b' neq_b e e') (at level 76) : abst_scope.
Notation "e >>= e'" := (bop_b' ge_b e e') (at level 76) : abst_scope.
Notation "e >> e'" := (bop_b' gt_b e e') (at level 76) : abst_scope.
Notation "e &e e'" := (bop_b2' and_b e e') (at level 77) : abst_scope.
Notation "e |e e'" := (bop_b2' or_b e e') (at level 77) : abst_scope.
Notation "e << e'" := (neg_b' (bop_b' ge_b e e')) (at level 76) : abst_scope.
Notation "e <<= e'" := (neg_b' (bop_b' gt_b e e')) (at level 76) : abst_scope.
Notation "~e e" := (neg_b' e) (at level 77) : abst_scope.

Fixpoint inj_expr_b e : expr_b' :=
  match e with
    | true_b => true_b'
    | bop_b b e1 e2 => bop_b' b (inj_expr e1) (inj_expr e2)
    | neg_b e1 => neg_b' (inj_expr_b e1)
    | bop_b2 b e1 e2 => bop_b2' b (inj_expr_b e1) (inj_expr_b e2)
  end.

Fixpoint expr_b'2expr_b e' : option expr_b :=
  match e' with
    | true_b' => Some true_b
    | bop_b' b e1' e2' => match (expr'2expr e1', expr'2expr e2') with
                            | (Some e1, Some e2) => Some (bop_b b e1 e2)
                            | _ => None
                          end
    | neg_b' e1' => match expr_b'2expr_b e1' with
                      | Some e1 => Some (neg_b e1)
                      | _ => None
                    end
    | bop_b2' b e1' e2' => match (expr_b'2expr_b e1', expr_b'2expr_b e2') with
                            | (Some e1, Some e2) => Some (bop_b2 b e1 e2)
                            | _ => None
                          end
  end.

Definition closed_b (e' : expr_b') :=
  match expr_b'2expr_b e' with Some _ => true | None => false end.

Lemma closed_b_bop_b' : forall b e1 e2, closed_b (bop_b' b e1 e2) ->
  closed e1 /\ closed e2.
Proof.
move=> b e1 e2.
rewrite /closed_b /= /closed.
set t1 := (expr'2expr e1).
set t2 := (expr'2expr e2).
destruct t1 => //.
by destruct t2.
Qed.

Definition eval_b' e' s := 
  match expr_b'2expr_b e' with 
    | Some e => Some (eval_b e s)
    | _ => None
  end.

Fixpoint subst_b' (e : expr_b') (patt repl : expr') : expr_b' :=
  match e with
    | true_b' => true_b'
    | bop_b' bo f g => 
      bop_b' bo (subst_e' f patt repl) (subst_e' g patt repl)
    | bop_b2' bo f g => 
      bop_b2' bo (subst_b' f patt repl) (subst_b' g patt repl)
    | neg_b' e => neg_b' (subst_b' e patt repl)
  end.  

Fixpoint expr_b'_fvar_set (e : expr_b') : list var.v :=
  match e with 
    | true_b' => nil
    | bop_b' _ e1 e2 => seq_ext.app_set (fun a b => a == b) (expr'_fvar_set e1) (expr'_fvar_set e2)
    | bop_b2' _ e1 e2 => seq_ext.app_set (fun a b => a == b) (expr_b'_fvar_set e1) (expr_b'_fvar_set e2)
    | ~e e => expr_b'_fvar_set e
  end.

(** separation logic formulas and properties *)

Inductive spatial : Type :=
| spa_mapsto : expr' -> expr' -> spatial
| spa_cell : expr' -> spatial
| spa_emp : spatial
| spa_ls : expr' -> expr' -> expr' -> spatial
| spa_tree : expr' -> expr' -> spatial.

Notation "e1 '|~>' e2" := (spa_mapsto e1 e2) (at level 77, no associativity) : abst_scope.
Notation " e '|~>_' " := (spa_cell e) (right associativity, at level 77) : abst_scope.

Definition spa_fvars (f : spatial) : list var.v :=
match f with
  | e1 |~> e2 => app_set (fun a b => a == b) (expr'_fvar_set e1) (expr'_fvar_set e2)
  | e |~>_ => expr'_fvar_set e
  | spa_emp => nil
  | spa_ls e1 e2 e3 => app_set (fun a b => a == b) (expr'_fvar_set e1) (app_set (fun a b => a == b) (expr'_fvar_set e2) (expr'_fvar_set e3))
  | spa_tree e1 e2 => app_set (fun a b => a == b) (expr'_fvar_set e1) (expr'_fvar_set e2)
end.

Definition spa_closed (f : spatial) : bool :=
  match f with
    | e1 |~> e2 => closed e1 && closed e2
    | e |~>_ => closed e
    | spa_emp => true
    | spa_ls e1 e2 e3 => closed e1 && closed e2 && closed e3
    | spa_tree e1 e2 => closed e1 && closed e2
  end.

Definition subst_spa (f : spatial) (x : nat) (e : expr') : spatial :=
  match f with
    | e1 |~> e2 => (subst_e' e1 (bvar_e x) e) |~> (subst_e' e2 (bvar_e x) e)
    | e' |~>_ => (subst_e' e' (bvar_e x) e) |~>_
    | spa_emp => spa_emp
    | spa_ls e1 e2 e3 => 
      spa_ls (subst_e' e1 (bvar_e x) e) (subst_e' e2 (bvar_e x) e) (subst_e' e3 (bvar_e x) e)
    | spa_tree e1 e2 => spa_tree (subst_e' e1 (bvar_e x) e) (subst_e' e2 (bvar_e x) e)
  end.

Inductive formula : Type :=
| for_bool : expr_b' -> formula
| for_spa : spatial -> formula
| for_con : formula -> formula -> formula
| for_and : formula -> formula -> formula
| for_or : formula -> formula -> formula
| for_exists : formula -> formula.

Notation "s ** t" := (for_con s t) (at level 80, right associativity) : abst_scope.
Notation "s //\\ t" := (for_and s t) (at level 80, right associativity) : abst_scope.
Notation "s \\// t" := (for_or s t) (at level 80, right associativity) : abst_scope.
Local Open Scope abst_scope.

Definition spatial_interp (p : spatial) : assert :=
  match p with
    | e1' |~> e2' => fun s h =>
      match (expr'2expr e1', expr'2expr e2') with
        | (Some e1, Some e2) => (e1 |~> e2)%seplog_assert s h /\ eval e1 s <> 0%Z
        | _ => False
      end
    | e' |~>_ => fun s h =>
      match expr'2expr e' with
        | Some e => exists e2, (e |~> e2)%seplog_assert s h /\ eval e s <> 0%Z
        | _ => False
      end
    | spa_emp => assert_m.emp
    | spa_ls n' e1' e2' =>
      match (expr'2expr n', expr'2expr e1', expr'2expr e2') with
        | (Some n, Some e1, Some e2) => ls n e1 e2
        | _ => FF
      end
    | spa_tree n' r' =>
      match (expr'2expr n', expr'2expr r') with
        | (Some n, Some r) => tree n r
        | _ => FF
      end
end.    

Fixpoint formula_fvars (f : formula) : list var.v :=
  match f with
    | for_bool eb => expr_b'_fvar_set eb
    | for_spa spa => spa_fvars spa
    | f1 ** f2 => app_set (fun a b => a == b) (formula_fvars f1) (formula_fvars f2)
    | f1 //\\ f2 => app_set (fun a b => a == b) (formula_fvars f1) (formula_fvars f2)
    | f1 \\// f2 => app_set (fun a b => a == b) (formula_fvars f1) (formula_fvars f2)
    | for_exists f' => formula_fvars f'
  end.

Fixpoint formula_closed (f : formula) : bool :=
  match f with
    | for_bool eb => closed_b eb
    | for_spa spa => spa_closed spa
    | f1 ** f2 => formula_closed f1 && formula_closed f2
    | f1 //\\ f2 => formula_closed f1 && formula_closed f2
    | f1 \\// f2 => formula_closed f1 && formula_closed f2
    | for_exists _ => false
  end.
      
Fixpoint subst_formula (f : formula) (i : nat) (e : expr') : formula :=
  match f with
    | for_bool eb => for_bool (subst_b' eb (bvar_e i) e)
    | for_spa f' => for_spa (subst_spa f' i e)
    | f1 ** f2 => subst_formula f1 i e ** subst_formula f2 i e
    | f1 //\\ f2 => subst_formula f1 i e //\\ subst_formula f2 i e
    | f1 \\// f2 => subst_formula f1 i e \\// subst_formula f2 i e
    | for_exists f1 => for_exists (subst_formula f1 (S i) e)
  end.
      
Fixpoint formula_size (f : formula) : nat :=
  match f with
    | f1 ** f2 => (1 + formula_size f1 + formula_size f2)%nat
    | f1 //\\ f2 => (1 + formula_size f1 + formula_size f2)%nat
    | f1 \\// f2 => (1 + formula_size f1 + formula_size f2)%nat
    | for_exists f' => (1 + formula_size f')%nat
    | for_bool _ => O
    | for_spa _ => O
  end.

Fixpoint formula_interp' (a : formula) (n : nat) : assert :=
  match n with
    | O => 
      match a with
        | for_bool e => 
          fun s _ => match eval_b' e s with Some x => x | _ => False end
        | for_spa p => spatial_interp p
        | _ => FF
      end
    | S n' =>
      match a with 
        | s1 ** s2 => 
          (formula_interp' s1 n' ** formula_interp' s2 n')%seplog_assert
        | s1 //\\ s2 => 
          (formula_interp' s1 n' //\\ formula_interp' s2 n')%seplog_assert
        | s1 \\// s2 => 
          (formula_interp' s1 n' \\// formula_interp' s2 n')%seplog_assert
        | for_exists sig => 
          fun s h => exists v, formula_interp' (subst_formula sig O (cst_e' v)) n' s h
        | for_bool e => 
          fun s _ => match eval_b' e s with Some x => x | _ => False end
        | for_spa p => 
          spatial_interp p
      end
  end.

Definition formula_interp a := formula_interp' a (formula_size a).

(* TODO: rename to formula_size_subst_formula_cst_e' *)
Lemma formula_size_subst_formula : forall f v n,
  formula_size (subst_formula f n (cst_e' v)) = formula_size f.
Proof.
elim=> //=.
move=> f1 IH1 f2 IH2 v n; by rewrite IH1 IH2.
move=> f1 IH1 f2 IH2 v n; by rewrite IH1 IH2.
move=> f1 IH1 f2 IH2 v n; by rewrite IH1 IH2.
move=> f IH v n; by rewrite IH.
Qed.

Lemma formula_interp'_formula_size : forall l f s h, (l >= formula_size f)%nat ->
  formula_interp' f l s h ->
  formula_interp' f (formula_size f) s h.
Proof.
move=> l f.
move Hn : (formula_size f) => n.
have {Hn}Hn : (formula_size f <= n)%nat by rewrite Hn.
move: f n Hn.
apply lt_wf_ind with (P := fun l => forall f n, (formula_size f <= n)%nat ->
  forall s h, (n <= l)%nat -> formula_interp' f l s h -> formula_interp' f n s h).
move=> {l} l IH [] //.
+ move=> e n /= Hn s h H; by destruct n; destruct l.
+ move=> f n /= Hn s h H; by destruct n; destruct l.
+ move=> f1 f2 n /= Hn s h H; destruct l => //; destruct n => //=.
  apply monotony => h' Hh'; apply IH with l => //; ssromega.
+ move=> f1 f2 n /= Hn s h H; destruct l => //; destruct n => //=.
  case=> H1 H2; split; apply IH with l => //; ssromega.
+ move=> f1 f2 n /= Hn s h H; destruct l => //; destruct n => //=.
  case=> X; [left | right]; apply IH with l => //; ssromega.
+ move=> f n /= Hn s h H X; destruct l => //; destruct n => //=.
  rewrite /= in X; case: X => v X; exists v.
  apply IH with l => //.
  rewrite formula_size_subst_formula; ssromega.
Qed.

Lemma formula_size_subst_formula_fvar_e : forall f x n,
  formula_size (subst_formula f n (fvar_e x)) = formula_size f.
Proof.
elim=> //=.
move=> f1 IH1 f2 IH2 x n; by rewrite IH1 IH2.
move=> f1 IH1 f2 IH2 x n; by rewrite IH1 IH2.
move=> f1 IH1 f2 IH2 x n; by rewrite IH1 IH2.
move=> f IH x n; by rewrite IH.
Qed.

Lemma subst_e'_bvar_e : forall e v x,
  ~ List.In x (expr'_fvar_set e) ->
  subst_e' e (bvar_e 0) (cst_e' v) =
  subst_e' (subst_e' e (bvar_e 0) (fvar_e x)) (fvar_e x) (cst_e' v).
Proof.
elim=> //.
- move=> x v y H.
  rewrite /= in H.
  have {H}H : x <> y by intuition.
  rewrite /=.
  case: ifP => //.
  move/eqP => ?; tauto.
- move=> n v x _.
  rewrite /=.
  case: ifP => // /eqP ?; subst n => /=.
  by rewrite eqxx.                                       
- move=> b e1 IH1 e2 IH2 v x /= H.
  rewrite -(IH1 v x); last first.
    contradict H.
    apply in_or_app_set; first by move=> *; apply/eqP.
    by auto.
  rewrite -(IH2 v x) //.
  contradict H.
  apply in_or_app_set; first by move=> *; apply/eqP.
  by auto.
- move=> uo e IH v x /= H.
  by rewrite -(IH v x).
Qed.

Fixpoint expr'_bvar_set (e : expr') : list nat :=
  match e with
    | fvar_e _ => nil
    | bvar_e x => x :: nil
    | cst_e' z => nil
    | bop_e' _ e1 e2 =>
      seq_ext.app_set (fun a b => a == b) (expr'_bvar_set e1) (expr'_bvar_set e2)
(*    | valabs_e' e' => expr'_bvar_set e'*)
    | uop_e' _ e' => expr'_bvar_set e'
  end.

Fixpoint expr_b'_bvar_set (e : expr_b') : list nat :=
  match e with 
    | true_b' => nil
    | bop_b' _ e1 e2 => seq_ext.app_set (fun a b => a == b) (expr'_bvar_set e1) (expr'_bvar_set e2)
    | bop_b2' _ e1 e2 => seq_ext.app_set (fun a b => a == b) (expr_b'_bvar_set e1) (expr_b'_bvar_set e2)
    | ~e e => expr_b'_bvar_set e
  end.

Definition spa_bvars (f : spatial) : list nat :=
match f with
  | e1 |~> e2 => app_set (fun a b => a == b) (expr'_bvar_set e1) (expr'_bvar_set e2)
  | e |~>_ => expr'_bvar_set e
  | spa_emp => nil
  | spa_ls e1 e2 e3 => app_set (fun a b => a == b) (expr'_bvar_set e1) (app_set beq_nat (expr'_bvar_set e2) (expr'_bvar_set e3))
  | spa_tree e1 e2 => app_set (fun a b => a == b) (expr'_bvar_set e1) (expr'_bvar_set e2)
end.

Fixpoint formula_bvars (f : formula) : list nat :=
  match f with
    | for_bool eb => expr_b'_bvar_set eb
    | for_spa spa => spa_bvars spa
    | f1 ** f2 => app_set (fun a b => a == b) (formula_bvars f1) (formula_bvars f2)
    | f1 //\\ f2 => app_set (fun a b => a == b) (formula_bvars f1) (formula_bvars f2)
    | f1 \\// f2 => app_set (fun a b => a == b) (formula_bvars f1) (formula_bvars f2)
    | for_exists f' => formula_bvars f'
  end.

Lemma eval'_None_inv : forall e s, eval' e s = None -> expr'2expr e = None.
Proof.
elim => //.
- move=> b e1 IH1 e2 IH2 s /=.
  move He1' : (expr'2expr e1) => e1'.
  destruct e1'; last by done.
  move He2' : (expr'2expr e2) => e2'.
  destruct e2'; last by done.
  by rewrite /eval' /= He1' He2'.
- move=> u e IH s /=.
  move He' : (expr'2expr e) => e'.
  destruct e'; last by done.
  by rewrite /eval' /= He'.
Qed.

Lemma eval'_bop_e'_inv : forall b e1 e2 s w,
  eval' (bop_e' b e1 e2) s = Some w ->
  exists v1, exists v2, eval' e1 s = Some v1 /\ eval' e2 s = Some v2.
Proof.
move=> b e1 e2 s w.
rewrite /eval' /=.
move He1' : (expr'2expr e1) => e1'.
destruct e1'; last by done.
move He2' : (expr'2expr e2) => e2'.
destruct e2'; last by done.
move=> Hw.
exists ([ e ]e_s). 
by exists ([ e0 ]e_s).
Qed.

Lemma eval'_uop_e'_inv : forall e uo w s,
  eval' (uop_e' uo e) s = Some w -> exists u, eval' e s = Some u.
Proof.
move=> e uo w s H.
have [X|X] : eval' e s = None \/ exists u : ZIT.t, eval' e s = Some u.
  destruct (eval' e s).
  right; by exists s0.
  by auto.
rewrite /eval' /= in H.
apply eval'_None_inv in X.
by rewrite X in H.
done.
Qed.

Lemma eval_b'_bop_b'_inv : forall b e1 e2 s w,
  eval_b' (bop_b' b e1 e2) s = Some w ->
  exists v1, exists v2, eval' e1 s = Some v1 /\ eval' e2 s = Some v2.
Proof.
move=> b e1 e2 s w.
rewrite /eval_b' /=.
move He1' : (expr'2expr e1) => e1'.
destruct e1'; last by done.
move He2' : (expr'2expr e2) => e2'.
destruct e2'; last by done.
move=> Hw.
rewrite /eval' /=.
rewrite He1' He2'.
exists ([ e ]e_s). 
by exists ([ e0 ]e_s).
Qed.

Lemma eval_b'_neg_b'_inv : forall b' w s, eval_b' (~e b') s = Some w ->
  exists v, eval_b' b' s = Some v.
Proof.
move=> b' w s /=.
rewrite /eval_b' /=.
move Hb'' : (expr_b'2expr_b _) => b''.
destruct b''; last by  done.
rewrite /= => H.
eexists.
reflexivity.
Qed.

Lemma eval_b'_bop_b2'_inv : forall b b1 b2 s w,
  eval_b' (bop_b2' b b1 b2) s = Some w ->
  exists v1, exists v2, eval_b' b1 s = Some v1 /\ eval_b' b2 s = Some v2.
Proof.
move=> b b1 b2 s w.
rewrite /eval_b' /=.
move Hb1' : (expr_b'2expr_b b1) => b1'.
destruct b1'; last by done.
move Hb2' : (expr_b'2expr_b b2) => b2'.
destruct b2'; last by done.
move=> Hw.
rewrite /eval' /=.
exists ([ e ]b_s). 
by exists ([ e0 ]b_s).
Qed.

Lemma incl_expr'_bvar_set : forall e v s w,
  eval' (subst_e' e (bvar_e 0) (cst_e' v)) s = Some w ->
  List.incl (expr'_bvar_set e) (O :: nil).
Proof.
elim.
- move=> x v s w /=.
  by intuition.
- move=> n v s w /=.
  case: ifP => // /eqP ?; subst n => /= _; done.
- move=> v w s u /=.
  by intuition.
- move=> b e1 IH1 e2 IH2 v s w /=.
  case/eval'_bop_e'_inv=> v1 [v2 []].
  move/IH1 => Hv1 /IH2 Hv2.
  apply incl_app_set => // *; by apply/eqP.
- move=> uo e IH v s w /=.
  case/eval'_uop_e'_inv => x.
  by move/IH.
Qed.

Lemma incl_expr_b'_bvar_set : forall b v s w,
  eval_b' (subst_b' b (bvar_e 0) (cst_e' v)) s = Some w ->
  List.incl (expr_b'_bvar_set b) (O :: nil).
Proof.
elim.
- move=> v s w /=.
  by intuition.
- move=> b e1 e2 v s w.
  rewrite /=.
  case/eval_b'_bop_b'_inv=> v1 [v2 []].
  move/incl_expr'_bvar_set => Hv1.
  move/incl_expr'_bvar_set => Hv2.
  apply incl_app_set => // *; by apply/eqP.
- move=> b IH v s w /=.
  case/eval_b'_neg_b'_inv=> v1 [].
  by move/IH.
- move=> b b1 IH1 b2 IH2 v s w /=.
  case/eval_b'_bop_b2'_inv => v1 [v2 []].
  move/IH1 => Hv1 /IH2 Hv2.
  apply incl_app_set => // *; by apply/eqP.
Qed.  

Lemma eval'_subst_e'_bvar_e_cst_e' : forall e v s x i,
  ~ List.In x (expr'_fvar_set e) ->
  eval' (subst_e' e (bvar_e i) (fvar_e x)) (store.upd x v s) = eval' (subst_e' e (bvar_e i) (cst_e' v)) s.
Proof.
elim => //.
- move=> y v s x i /= Hx.
  have {Hx}Hx : y <> x. by intuition.
  rewrite /eval'.
  rewrite /expr'2expr /=.
  by rewrite store.get_upd.
- move=> y v s z i /= _.
  case: ifP => X //.
  move/eqP in X; subst y.
  rewrite /eval' /=.
  by rewrite store.get_upd'.
- move=> b e1 IH1 e2 IH2 v s x i /= Hx.
  have H1 : ~ List.In x (expr'_fvar_set e1).
    contradict Hx.
    apply in_or_app_set; first by move=> *; apply/eqP.
    by auto.
  have H2 : ~ List.In x (expr'_fvar_set e2).
    contradict Hx.
    apply in_or_app_set; first by move=> *; apply/eqP.
    by auto.
  move: {H1}(IH1 v s x i H1) => H1.
  move: {H2}(IH2 v s x i H2) => H2.
  move: H1 H2.
  rewrite /eval' /=.
  move He1' : (expr'2expr (subst_e' e1 _ _)) => e1'.
  destruct e1' => //.
  move He2' : (expr'2expr (subst_e' e2 _ _)) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr (subst_e' e2 _ _)) => e2''.
  destruct e2'' => //=.
  by case=> <- [<-].
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr (subst_e' e2 _ _)) => e2''.
  by destruct e2''.
  move He2' : (expr'2expr (subst_e' e2 _ _)) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  by destruct e1''.
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  by destruct e1''.
- move=> uo e IH v s x i Hx.
  have Hx' : ~ List.In x (expr'_fvar_set e).
    by contradict Hx.
  move: {IH}(IH v s x i Hx').
  rewrite /eval' /=.
  move He1 : (expr'2expr (subst_e' e _ _)) => e1.
  destruct e1 => //.
  move He1' : (expr'2expr (subst_e' e _ _)) => e1'.
  destruct e1' => //=.
  by case => <-.
  move He2 : (expr'2expr (subst_e' e _ _)) => e2.
  by destruct e2.
Qed.

Lemma eval_b'_subst_b'_bvar_e_cst_e' : forall b v s x i,
  ~ List.In x (expr_b'_fvar_set b) ->
  eval_b' (subst_b' b (bvar_e i) (fvar_e x)) (store.upd x v s) = eval_b' (subst_b' b (bvar_e i) (cst_e' v)) s.
Proof.
elim=> //.
- move=> b e1 e2 v s x i /= Hx.
  have H1 : ~ List.In x (expr'_fvar_set e1).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have H2 : ~ List.In x (expr'_fvar_set e2).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  move: {H1}(eval'_subst_e'_bvar_e_cst_e' e1 v s x i H1) => H1.
  move: {H2}(eval'_subst_e'_bvar_e_cst_e' e2 v s x i H2) => H2.
  move: H1 H2.
  rewrite /eval' /eval_b' /=.
  move He1' : (expr'2expr (subst_e' e1 _ _)) => e1'.
  destruct e1' => //.
  move He2' : (expr'2expr (subst_e' e2 _ _)) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr (subst_e' e2 _ _)) => e2''.
  destruct e2'' => //=.
  by case=> <- [<-].
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr (subst_e' e2 _ _)) => e2''.
  by destruct e2''.
  move He2' : (expr'2expr (subst_e' e2 _ _)) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  by destruct e1''.
  move He1'' : (expr'2expr (subst_e' e1 _ _)) => e1''.
  by destruct e1''.
- move=> e IH v s x i /= Hx.
  move: {Hx}(IH v s x i Hx).
  rewrite /eval_b' /=.
  move He' : (expr_b'2expr_b (subst_b' e _ _)) => e'.
  destruct e' => //.
  move He'' : (expr_b'2expr_b (subst_b' e _ _)) => e''.
  destruct e'' => //=.
  by case=> <-.
  move He'' : (expr_b'2expr_b (subst_b' e _ _)) => e''.
  by destruct e''.
- move=> b e1 IH1 e2 IH2 v s x i /= Hx.
  have H1 : ~ List.In x (expr_b'_fvar_set e1).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have H2 : ~ List.In x (expr_b'_fvar_set e2).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  move: {H1}(IH1 v s x i H1) => H1.
  move: {H2}(IH2 v s x i H2) => H2.
  move: H1 H2.
  rewrite /eval_b' /=.
  move He1' : (expr_b'2expr_b (subst_b' e1 _ _)) => e1'.
  destruct e1' => //.
  move He2' : (expr_b'2expr_b (subst_b' e2 _ _)) => e2'.
  destruct e2' => //.
  move He1'' : (expr_b'2expr_b (subst_b' e1 _ _)) => e1''.
  destruct e1'' => //.
  move He2'' : (expr_b'2expr_b (subst_b' e2 _ _)) => e2''.
  destruct e2'' => //=.
  by case=> <- [<-].
  move He1'' : (expr_b'2expr_b (subst_b' e1 _ _)) => e1''.
  destruct e1'' => //.
  move He2'' : (expr_b'2expr_b (subst_b' e2 _ _)) => e2''.
  by destruct e2''.
  move He2' : (expr_b'2expr_b (subst_b' e2 _ _)) => e2'.
  destruct e2' => //.
  move He1'' : (expr_b'2expr_b (subst_b' e1 _ _)) => e1''.
  by destruct e1''.
  move He1'' : (expr_b'2expr_b (subst_b' e1 _ _)) => e1''.
  by destruct e1''.
Qed.

Lemma formula_interp'_subst_formula_spa : forall f s h x v i,
  ~ List.In x (formula_fvars (for_spa f)) ->
  formula_interp' (subst_formula (for_spa f) i (cst_e' v)) O s h ->
  formula_interp' (subst_formula (for_spa f) i (fvar_e x)) O (store.upd x v s) h.
Proof.
elim => //.
- move=> e1 e2 s h x v i Hx /=.
  rewrite /= in Hx.
  have He1 : ~ List.In x (expr'_fvar_set e1).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have He2 : ~ List.In x (expr'_fvar_set e2).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  move: {He1}(eval'_subst_e'_bvar_e_cst_e' e1 v s x i He1).
  move: {He2}(eval'_subst_e'_bvar_e_cst_e' e2 v s x i He2).
  rewrite /eval' /=.
  move He1' : (expr'2expr _) => e1'.
  destruct e1' => //.
  move He2' : (expr'2expr _) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr _) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr _) => e2''.
  destruct e2'' => //.
  case=> H1 [H2].
  case=> H3 H4.
  split.
  + move: H3; by apply mapsto_ext.
  + contradict H4.
    congruence.
  move He2'' : (expr'2expr _) => e2''.
  by destruct e2''.
  move He2' : (expr'2expr _) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr _) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr _) => e2''.
  by destruct e2''.
  move He2'' : (expr'2expr _) => e2''.
  by destruct e2''.
- move=> e s h x v i Hx /=.
  rewrite /= in Hx.
  move: {Hx}(eval'_subst_e'_bvar_e_cst_e' e v s x i Hx).
  rewrite /eval' /=.
  move He' : (expr'2expr _) => e'.
  destruct e' => //.
  move He'' : (expr'2expr _) => e''.
  destruct e'' => //.
  case=> H1 [e2 [He1e2 He2]].
  exists (cst_e [ e2 ]e_s).
  split.
  + move: He1e2; by eapply mapsto_ext.
  + contradict He2.
    congruence.
  move He'' : (expr'2expr _) => e''.
  by destruct e''.
- move=> e1 e2 e3 s h x v i Hx.
  rewrite /= in Hx.
  have He1 : ~ List.In x (expr'_fvar_set e1).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have He2 : ~ List.In x (expr'_fvar_set e2).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP.
    right.
    apply in_or_app_set; first by move=> *; apply/eqP. by auto. 
  have He3 : ~ List.In x (expr'_fvar_set e3).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP.
    right.
    apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  move: {He1}(eval'_subst_e'_bvar_e_cst_e' e1 v s x i He1).
  move: {He2}(eval'_subst_e'_bvar_e_cst_e' e2 v s x i He2).
  move: {He3}(eval'_subst_e'_bvar_e_cst_e' e3 v s x i He3).
  rewrite /eval' /=.
  move He3' : (expr'2expr _) => e3'.
  destruct e3' => //.
  move He3'' : (expr'2expr _) => e3''.
  destruct e3'' => //.
  move He2' : (expr'2expr _) => e2'.
  destruct e2' => //.
  move He2'' : (expr'2expr _) => e2''.
  destruct e2'' => //.
  move He1' : (expr'2expr _) => e1'.
  destruct e1' => //.
  move He1'' : (expr'2expr _) => e1''.
  destruct e1'' => //.
  case=> X [Y] [Z] H.
  eapply ls_equiv; first by apply H.
  done.
  done.
  done.
  move He1'' : (expr'2expr _) => e1''.
  by destruct e1''.
  move He2'' : (expr'2expr _) => e2''.
  destruct e2'' => //.
  move He1'' : (expr'2expr _) => e1''.
  destruct e1'' => //.
  move He1''' : (expr'2expr _) => e1'''.
  by destruct e1'''.
  move He1'''' : (expr'2expr _) => e1''''.
  by destruct e1''''.
  move He3'' : (expr'2expr _) => e3''.
  destruct e3'' => //.
  move He2' : (expr'2expr _) => e2'.
  destruct e2' => //.
  move He2'' : (expr'2expr _) => e2''.
  destruct e2'' => //.
  move He1' : (expr'2expr _) => e1'.
  destruct e1' => //.
  move He1'' : (expr'2expr _) => e1''.
  by destruct e1''.
  move He1'' : (expr'2expr _) => e1''.
  by destruct e1''.
  move He2'' : (expr'2expr _) => e2''.
  destruct e2'' => //.
  move He1' : (expr'2expr _) => e1'.
  destruct e1' => //.
  move He1'' : (expr'2expr _) => e1''.
  by destruct e1''.
  move He1'' : (expr'2expr _) => e1''.
  by destruct e1''.
- move=> e1 e2 s h x v i Hx /=.
  rewrite /= in Hx.
  have He1 : ~ List.In x (expr'_fvar_set e1).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have He2 : ~ List.In x (expr'_fvar_set e2).
    contradict Hx. apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  move: {He1}(eval'_subst_e'_bvar_e_cst_e' e1 v s x i He1).
  move: {He2}(eval'_subst_e'_bvar_e_cst_e' e2 v s x i He2).
  rewrite /eval' /=.
  move He1' : (expr'2expr _) => e1'.
  destruct e1' => //.
  move He2' : (expr'2expr _) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr _) => e1''.
  destruct e1'' => //.
  move He2'' : (expr'2expr _) => e2''.
  destruct e2'' => //.
  case=> H1 [H2] H3.
  eapply tree_equiv; first by apply H3.
  done.
  done.
  move He1''' : (expr'2expr _) => e1'''.
  by destruct e1'''.
  move He2' : (expr'2expr _) => e2'.
  destruct e2' => //.
  move He1'' : (expr'2expr _) => e1''.
  destruct e1'' => //.
  move He1''' : (expr'2expr _) => e1'''.
  by destruct e1'''.
  move He1''' : (expr'2expr _) => e1'''.
  by destruct e1'''.
Qed.

Lemma formula_interp'_formula_size2 : forall l f s h, 
  (l >= formula_size f)%nat ->
  formula_interp' f (formula_size f) s h ->
  formula_interp' f l s h.
Proof.
move=> l f.
move Hn : (formula_size f) => n.
have {Hn}Hn : (formula_size f <= n)%nat by rewrite Hn.
move: f n Hn.
apply lt_wf_ind with (P := fun l => forall (f : formula) (n : nat),
   (formula_size f <= n)%nat ->
   forall (s : store.t) (h : heap.t),
   (n <= l)%nat -> formula_interp' f n s h -> formula_interp' f l s h).
move=> {l} l IH [] //.
+ move=> e n /= Hn s h Hl; by destruct n; destruct l.
+ move=> f n /= Hn s h Hl; by destruct n; destruct l.
+ move=> f1 f2 n /= Hn s h Hl; destruct n => //; destruct l => //=.
  apply monotony => h' Hh'; apply IH with l => //.
  ssromega.
  apply IH with n => //; ssromega.
  ssromega.
  apply IH with n => //; ssromega.
+ move=> f1 f2 n /= Hn s h Hl; destruct l => //; destruct n => //=.
  case=> H1 H2; split; apply IH with l => //.
  ssromega.
  apply IH with n => //; ssromega.
  ssromega.
  apply IH with n => //; ssromega.
+ move=> f1 f2 n /= Hn s h Hl; destruct l => //; destruct n => //=.
  case=> X; [left | right]; apply IH with l => //.
  ssromega.
  apply IH with n => //; ssromega.
  ssromega.
  apply IH with n => //; ssromega.
+ move=> f n /= Hn s h Hl X; destruct l => //; destruct n => //=.
  rewrite /= in X; case: X => v X; exists v.
  apply IH with n => //.
  rewrite formula_size_subst_formula.
  ssromega.
Qed.

Lemma In_expr'_fvar_set_inv : forall e x n v,
  List.In x (expr'_fvar_set (subst_e' e (bvar_e n) (cst_e' v))) ->
  List.In x (expr'_fvar_set e).
Proof.
elim => //.
- move=> n x n0 v /=.
  by case: ifP.
- move=> b e1 IH1 e2 IH2 x n v /= Hx.
  apply in_app_set_or in Hx.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  eapply IH1; eauto.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  eapply IH2; eauto. by move=> *; apply/eqP.
Qed.

Lemma In_expr_b'_fvar_set_inv : forall e x v n,
  List.In x (expr_b'_fvar_set (subst_b' e (bvar_e n) (cst_e' v))) ->
  List.In x (expr_b'_fvar_set e).
Proof.
elim=> //.
- move=> b e1 e2 x v n /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by apply In_expr'_fvar_set_inv in Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by apply In_expr'_fvar_set_inv in Hx.
- move=> b e1 IH1 e2 IH2 x v n /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  eapply IH1; eauto.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  eapply IH2; eauto.
Qed.

Lemma In_spa_fvars_inv : forall f x n v, 
  List.In x (spa_fvars (subst_spa f n (cst_e' v))) -> List.In x (spa_fvars f).
Proof.
elim=> //.
- move=> e1 e2 x n v /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by apply In_expr'_fvar_set_inv in Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by apply In_expr'_fvar_set_inv in Hx.
- move=> e x n v /= Hx.
  by apply In_expr'_fvar_set_inv in Hx.
- move=> e1 e2 e3 x n v /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by apply In_expr'_fvar_set_inv in Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by apply In_expr'_fvar_set_inv in Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by apply In_expr'_fvar_set_inv in Hx.
- move=> e1 e2 x n v /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by apply In_expr'_fvar_set_inv in Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by apply In_expr'_fvar_set_inv in Hx.
Qed.

Lemma In_formula_fvars_subst_formula_inv : forall f x v n,
  List.In x (formula_fvars (subst_formula f n (cst_e' v))) ->
  List.In x (formula_fvars f).
Proof.
elim => //.
- move=> e x v n /= Hx.
  by apply In_expr_b'_fvar_set_inv in Hx.
- move=> f x v n /= Hx.
  by apply In_spa_fvars_inv in Hx.
- move=> f1 IH1 f2 IH2 x v n /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by eapply IH1; eauto.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by eapply IH2; eauto.
- move=> f1 IH1 f2 IH2 x v n /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by eapply IH1; eauto.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by eapply IH2; eauto.
- move=> f1 IH1 f2 IH2 x v n /= Hx.
  apply in_app_set_or in Hx; last by move=> *; apply/eqP.
  case: Hx => Hx.
  apply in_or_app_set; first by move=> *; apply/eqP.
  left.
  by eapply IH1; eauto.
  apply in_or_app_set; first by move=> *; apply/eqP.
  right.
  by eapply IH2; eauto.
- move=> f IH x v n /= Hx.
  eapply IH; eauto.
Qed.


Lemma subst_e'_swap : forall e i j x v,
  i <> j ->
  subst_e' (subst_e' e (bvar_e i) (fvar_e x)) (bvar_e j) (cst_e' v) =
  subst_e' (subst_e' e (bvar_e j) (cst_e' v)) (bvar_e i) (fvar_e x).
Proof.
elim => //.
- move=> n i j x v Hij /=.
  case: ifP.
    move/eqP => ?; subst n.
    move/eqP/negbTE : Hij => -> /=.
    by rewrite eqxx.
  move=> X.
  case: ifP.
    move/eqP => ?; subst j => /=; by rewrite eqxx.
  move=> Y.
  by rewrite /= X Y.
- move=> b e1 IH1 e2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> uo e IH i j x v Hij /=.
  by rewrite IH.
Qed.

Lemma subst_b'_swap : forall e i j x v,
  i <> j ->
  subst_b' (subst_b' e (bvar_e i) (fvar_e x)) (bvar_e j) (cst_e' v) =
  subst_b' (subst_b' e (bvar_e j) (cst_e' v)) (bvar_e i) (fvar_e x).
Proof.
elim => //.
- move=> b e1 e2 i j x v Hij /=.
  by rewrite subst_e'_swap // subst_e'_swap.
- move=> e IH i j x v Hij /=.
  by rewrite IH.
- move=> b e1 IH1 e2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
Qed.

Lemma subst_spa_swap : forall f i j x v,
  i <> j ->
  for_spa (subst_spa (subst_spa f i (fvar_e x)) j (cst_e' v)) =
  for_spa (subst_spa (subst_spa f j (cst_e' v)) i (fvar_e x)).
Proof.
elim => //.
- move=> e1 e2 i j x v Hij /=.
  by rewrite subst_e'_swap // subst_e'_swap.
- move=> e i j x v Hij /=.
  by rewrite subst_e'_swap.
- move=> e1 e2 e3 i j x v Hij /=.
  by rewrite subst_e'_swap // subst_e'_swap // subst_e'_swap .
- move=> e1 e2 i j x v Hij /=.
  by rewrite subst_e'_swap // subst_e'_swap.
Qed.

Lemma subst_formula_swap : forall f i j x v,
  i <> j ->
  subst_formula (subst_formula f i (fvar_e x)) j (cst_e' v) =
  subst_formula (subst_formula f j (cst_e' v)) i (fvar_e x).
Proof.
elim => //.
- move=> e i j x v Hij /=.
  by rewrite subst_b'_swap.
- move=> f i j x v Hij /=.
  by rewrite subst_spa_swap.
- move=> f1 IH1 f2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> f1 IH1 f2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> f1 IH1 f2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> f IH i j Hij x v /=.
  rewrite IH //.
  by case.
Qed.

Lemma subst_e'_swap2 : forall e i j x v,
  i <> j ->
  subst_e' (subst_e' e (bvar_e i) (cst_e' x)) (bvar_e j) (cst_e' v) =
  subst_e' (subst_e' e (bvar_e j) (cst_e' v)) (bvar_e i) (cst_e' x).
Proof.
elim => //.
- move=> n i j x v Hij /=.
  case: ifP.
    move/eqP => ?; subst i => /=.
    case: ifP.
      move/eqP => ?; subst j; tauto.
    move=> Y /=; by rewrite eqxx.
  case: ifP => //.
    move/eqP => ?; subst j => /=.                   
    by rewrite eqxx.
  by rewrite /= => -> ->.
- move=> b e1 IH1 e2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> uo e IH i j x v Hij /=.
  by rewrite IH.
Qed.

Lemma subst_b'_swap2 : forall e i j x v,
  i <> j ->
  subst_b' (subst_b' e (bvar_e i) (cst_e' x)) (bvar_e j) (cst_e' v) =
  subst_b' (subst_b' e (bvar_e j) (cst_e' v)) (bvar_e i) (cst_e' x).
Proof.
elim => //.
- move=> b e1 e2 i j x v Hij /=.
  by rewrite (subst_e'_swap2 _ i j) // (subst_e'_swap2 _ i j).
- move=> e IH i j x v Hij /=.
  by rewrite IH.
- move=> b e1 IH1 e2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
Qed.

Lemma subst_spa_swap2 : forall f i j x v, i <> j ->
  for_spa (subst_spa (subst_spa f i (cst_e' x)) j (cst_e' v)) =
  for_spa (subst_spa (subst_spa f j (cst_e' v)) i (cst_e' x)).
Proof.
elim => //.
- move=> e1 e2 i j x v Hij /=.
  by rewrite (subst_e'_swap2 _ i j) // (subst_e'_swap2 _ i j).
- move=> e i j x v Hij /=.
  by rewrite subst_e'_swap2.
- move=> e1 e2 e3 i j x v Hij /=.
  by rewrite (subst_e'_swap2 _ i j) // (subst_e'_swap2 _ i j) // (subst_e'_swap2 _ i j).
- move=> e1 e2 i j x v Hij /=.
  by rewrite (subst_e'_swap2 _ i j) // (subst_e'_swap2 _ i j).
Qed.

Lemma subst_formula_swap2 : forall f i j x v,
  i <> j ->
  subst_formula (subst_formula f i (cst_e' x)) j (cst_e' v) =
  subst_formula (subst_formula f j (cst_e' v)) i (cst_e' x).
Proof.
elim => //.
- move=> e i j x v Hij /=.
  by rewrite subst_b'_swap2.
- move=> f i j x v Hij /=.
  by rewrite subst_spa_swap2.
- move=> f1 IH1 f2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> f1 IH1 f2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> f1 IH1 f2 IH2 i j x v Hij /=.
  by rewrite IH1 // IH2.
- move=> f IH i j Hij x v /=.
  rewrite IH //.
  by case.
Qed.

Lemma formula_interp'_subst_formula n : forall Q, 
  (formula_size Q = n)%nat ->
  forall s h x v i, ~ List.In x (formula_fvars Q) ->
  formula_interp' (subst_formula Q i (cst_e' v)) (formula_size Q) s h ->
  formula_interp' (subst_formula Q i (fvar_e x)) (formula_size Q) (store.upd x v s) h.
Proof.
apply lt_wf_ind with (P := fun n => forall Q,
   formula_size Q = n -> forall s h x v i,
   ~ List.In x (formula_fvars Q) ->
   formula_interp' (subst_formula Q i (cst_e' v)) (formula_size Q) s h ->
   formula_interp' (subst_formula Q i (fvar_e x)) 
     (formula_size Q) (store.upd x v s) h).
move=> {n} n IH.
case=> //.
+ move=> b _ s h x v i /= Hx.
  by rewrite eval_b'_subst_b'_bvar_e_cst_e'.
+ move=> f _ s h x v Hx.
  by apply formula_interp'_subst_formula_spa.
+ move=> f1 f2 /= Hn s h x v i Hx.
  apply monotony => h' Hh'.
  * apply formula_interp'_formula_size in Hh'; last first.
      rewrite formula_size_subst_formula; rewrite /addn_rec; ssromega.
    rewrite formula_size_subst_formula in Hh'.
    have H1 : ~ List.In x (formula_fvars f1).
      contradict Hx; apply in_or_app_set; first by move=> *; apply/eqP. by auto.
    have Htmp : (formula_size f1 < n)%coq_nat by ssromega.
    move: {IH}(IH _ Htmp f1 (refl_equal _) s h' x v i H1 Hh') => IH.
    apply formula_interp'_formula_size2.
    rewrite formula_size_subst_formula_fvar_e /addn_rec; ssromega.
    by rewrite formula_size_subst_formula_fvar_e.
  * apply formula_interp'_formula_size in Hh'; last first.
      rewrite formula_size_subst_formula /addn_rec; ssromega.
    rewrite formula_size_subst_formula in Hh'.
    have H2 : ~ List.In x (formula_fvars f2).
      contradict Hx; apply in_or_app_set; first by move=> *; apply/eqP. by auto.
    have Htmp : (formula_size f2 < n)%coq_nat by ssromega.
    move: {IH}(IH _ Htmp f2 (refl_equal _) s h' x v i H2 Hh') => IH.
    apply formula_interp'_formula_size2.
    rewrite formula_size_subst_formula_fvar_e /addn_rec; ssromega.
    by rewrite formula_size_subst_formula_fvar_e.
+ move=> f1 f2 /= Hn s h x v i Hx [H1 H2].
  apply formula_interp'_formula_size in H1; last first.
    rewrite formula_size_subst_formula /addn_rec; ssromega.
  rewrite formula_size_subst_formula in H1.
  apply formula_interp'_formula_size in H2; last first.
    rewrite formula_size_subst_formula /addn_rec; ssromega.
  rewrite formula_size_subst_formula in H2.
  have H1' : ~ List.In x (formula_fvars f1).
    contradict Hx; apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have H2' : ~ List.In x (formula_fvars f2).
    contradict Hx; apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have Htmp1 : (formula_size f1 < n)%coq_nat by ssromega.
  have Htmp2 : (formula_size f2 < n)%coq_nat by ssromega.
  move: (IH _ Htmp1 f1 (refl_equal _) s h x v i H1' H1) => IH1.
  move: {IH}(IH _ Htmp2 f2 (refl_equal _) s h x v i H2' H2) => IH2.
  split.
  * apply formula_interp'_formula_size2.
    rewrite formula_size_subst_formula_fvar_e /addn_rec; ssromega.
    by rewrite formula_size_subst_formula_fvar_e.
  * apply formula_interp'_formula_size2.
    rewrite formula_size_subst_formula_fvar_e /addn_rec; ssromega.
    by rewrite formula_size_subst_formula_fvar_e.
+ move=> f1 f2 /= Hn s h x v i Hx [H1 | H2].
  apply formula_interp'_formula_size in H1; last first.
    rewrite formula_size_subst_formula /addn_rec; ssromega.
  rewrite formula_size_subst_formula in H1.
  have H1' : ~ List.In x (formula_fvars f1).
    contradict Hx; apply in_or_app_set; first by move=> *; apply/eqP. by auto.
  have Htmp1 : (formula_size f1 < n)%coq_nat.
    rewrite -!plusE in Hn; omega.
  move: {IH}(IH _ Htmp1 f1 (refl_equal _) s h x v i H1' H1) => IH1.
  left.
  apply formula_interp'_formula_size2.
  rewrite formula_size_subst_formula_fvar_e /addn_rec; ssromega.
  by rewrite formula_size_subst_formula_fvar_e.
  * apply formula_interp'_formula_size in H2; last first.
      rewrite formula_size_subst_formula /addn_rec; ssromega.
    rewrite formula_size_subst_formula in H2.
    have H2' : ~ List.In x (formula_fvars f2).
      contradict Hx; apply in_or_app_set; first by move=> *; apply/eqP. by auto.
    have Htmp2 : (formula_size f2 < n)%coq_nat by ssromega.
    move: {IH}(IH _ Htmp2 f2 (refl_equal _) s h x v i H2' H2) => IH2.
    right.
    apply formula_interp'_formula_size2.
    rewrite formula_size_subst_formula_fvar_e /addn_rec; ssromega.
    by rewrite formula_size_subst_formula_fvar_e.
+ move=> f /= Hn s h x v i Hx [v0 Hv0].
  exists v0.
  rewrite (_ : (subst_formula (subst_formula f i.+1 (fvar_e x)) 0 (cst_e' v0)) =
    (subst_formula (subst_formula f 0 (cst_e' v0)) i.+1 (fvar_e x))); last first.
    by apply subst_formula_swap.
  rewrite (_ : formula_size f = formula_size ((subst_formula f 0 (cst_e' v0)))); last first.
    by rewrite formula_size_subst_formula.
  apply IH with (formula_size f).
  ssromega.
  by rewrite formula_size_subst_formula.
  contradict Hx.
  by apply In_formula_fvars_subst_formula_inv in Hx.
  rewrite subst_formula_swap2 //.
  by rewrite formula_size_subst_formula.
Qed.

Lemma formula_interp_subst_formula s h Q x v : ~ List.In x (formula_fvars Q) ->
  formula_interp (subst_formula Q 0 (cst_e' v)) s h ->
  formula_interp (subst_formula Q 0 (fvar_e x)) (store.upd x v s) h.
Proof.
move=> HxQ H.
rewrite /formula_interp formula_size_subst_formula in H.
rewrite /formula_interp formula_size_subst_formula_fvar_e.
by apply formula_interp'_subst_formula with (formula_size Q).
Qed.

Lemma tree_unroll n r :
  formula_interp (for_spa (spa_tree (fvar_e n) (fvar_e r))) ===>
  formula_interp (
    for_bool (fvar_e n =e cst_e' 0 &e fvar_e r =e cst_e' 0) \\//
    (for_bool (fvar_e n >> cst_e' 0) //\\
      for_exists (for_exists (
        for_bool (bvar_e 1 << fvar_e n &e bvar_e 0 << fvar_e n) //\\
        for_exists (for_exists (
          for_spa (fvar_e r |~>_) **
          for_spa (r -.> left |~> bvar_e 1) **
          for_spa (r -.> right |~> bvar_e 0) **
          for_spa (spa_tree (bvar_e 3) (bvar_e 1)) **
          for_spa (spa_tree (bvar_e 2) (bvar_e 0)))))))).
Proof.
move=> s h.
rewrite /formula_interp /= => H.
inversion H; subst; [left | right].
- rewrite /= in H1 H2; by rewrite H1 H2.
- split.
  + rewrite /= in H0; by omegab.
  + exists ([ n1 ]e_ s); exists ([ n2 ]e_ s).
    split.
    * apply/andP; split; omegab.
    * exists ([ lc ]e_s); exists ([ rc ]e_s).
      rewrite -!conAE.
      Compose_sepcon (h1 \U h2) (h3); last by eapply tree_equiv; eauto.
      Compose_sepcon h1 h2; last by eapply tree_equiv; eauto.
      rewrite -!conAE in H7.
      case_sepcon H7.
      case_sepcon H7_h11.
      Compose_sepcon (h111 \U h112) h12.
      - apply con_cons => //.
        case: H7_h11_h111 => e2 H7_h11_h111.
        exists e2.
        split; first by done.
        admit. (* see the definition of tree *)
        split; first by done.
        admit.
      - split; first by done.
        admit.
Admitted.

(** proof system of Figure 11 *)

Definition Gamma_type := list (nat * formula).

Require OrderedTypeEx FSetList.

Module varset_m := FSetList.Make (OrderedTypeEx.Nat_as_OT).

Local Open Scope seplog_hoare_scope.

Inductive cont_hoare : formula -> cont -> formula -> Prop :=
| hoare_def : forall P c Q, 
  {{ formula_interp P }} while.cmd_cmd0 c {{ formula_interp Q }} ->
  cont_hoare P (c ; cont_halt) Q.

Reserved Notation " G |- Q -| k' <| V |> k" (at level 82, no associativity).

Inductive abst : Gamma_type -> formula(*precondition*) -> cont -> varset_m.t -> cont -> Prop :=
| abst_halt : forall G Q V,
  G |- Q -| cont_halt <| V |> cont_halt

| abst_goto : forall G l V Q,
  assoc_get l G = Some Q ->
  G |- Q -| cont_goto l <| V |> cont_goto l

| abst_command : forall G Q c k_hat V k Q',
  cont_hoare Q (c; cont_halt) Q' ->
  G |- Q' -| k_hat <| V |> k ->
  G |- Q -| c ; k_hat <| V |> c ; k
| abst_branch : forall G Q es k_hats V ks,

  length es = length k_hats -> length k_hats = length ks ->
   (forall i, (i < length ks)%nat(*NB: nth defaults below will never be used*) -> 
     G |- Q //\\ for_bool (inj_expr_b (nth true_b es i)) -| 
       nth cont_abort k_hats i <| V |> nth cont_abort ks i) -> 
   G |- Q -| cont_branch es k_hats <| V |> cont_branch es ks 

| abst_inst_assign : forall G Q x e k_hat V k Q',
  varset_m.mem x V ->
  cont_hoare Q (x <- e; cont_halt) Q' ->
  abst G Q' k_hat V k ->
  G |- Q -| (x <- e; k_hat) <| V |> k

| abst_inst_assume : forall G Q eb k_hat V k,
  formula_interp Q ===> formula_interp (for_bool (inj_expr_b eb)) ->
  G |- Q -| k_hat <| V |> k ->
  G |- Q -| cont_branch (eb :: nil) (k_hat :: nil) <| V |> k

| abst_inst_exists : forall G x Q k_hat V k x' L,
  ~ List.In x' L ->
  varset_m.mem x V ->
  G |- subst_formula Q O (fvar_e x') -| k_hat <| V |> k ->
  G |- for_exists Q -| (x <-?; k_hat) <| V |> k

| abst_stren : forall G Q p_hat V p Q',
  formula_interp Q ===> formula_interp Q' ->
  G |- Q' -| p_hat <| V |> p ->
  G |- Q -| p_hat <| V |> p

where "G |- Q -| k' <| V |> k" := (abst G Q k' V k) : abst_scope.

(** first example of section 5.3 *)

Definition ls_prg x := (cont_branch
  ((var_e x \!= NIL) :: (var_e x \= NIL) :: nil)
  ((x <-* x -.> next ; cont_goto 1) :: cont_halt :: nil))%seplog_expr.

Definition instrumented_ls_prg (x n1 n2 : var.v) := (cont_branch
  ((var_e x \!= NIL) :: (var_e x \= NIL) :: nil)
  ((x <-* x -.> next ; 
    n1 <- var_e n1 \+ cst_e 1 ; 
    n2 <- var_e n2 \- cst_e 1 ; 
    cont_goto 1) :: cont_halt :: nil))%seplog_expr.

Definition ls_Gamma_1 (x : var.v) (n1 n2 : expr') : formula := 
  for_exists 
  (for_spa (spa_ls n1 (bvar_e O) (fvar_e x)) ** for_spa (spa_ls n2 (fvar_e x) (inj_expr NIL))).

Lemma example_53a : forall (x n1 n2 : var.v), 
  uniq (x :: n1 :: n2 :: nil) ->
  (1%nat, ls_Gamma_1 x (fvar_e n1) (fvar_e n2)) :: nil |- ls_Gamma_1 x (fvar_e n1) (fvar_e n2) -| 
    instrumented_ls_prg x n1 n2 <| varset_m.add n1 (varset_m.singleton n2) |> ls_prg x.
Proof.
move=> x n1 n2 Hset.
rewrite /ls_prg /instrumented_ls_prg.
apply abst_branch => // i Hi.
destruct i.
- move=> {Hi} /=.
  apply abst_command with 
    (ls_Gamma_1 x (fvar_e n1 \+ cst_e' 1) (fvar_e n2 \- cst_e' 1)).
  + apply hoare_def.
    apply hoare_lookup'.
    rewrite /wp_lookup => s h.
    rewrite /formula_interp [formula_size _]/= [formula_interp' _ _]/=.
    move=> [[v H] Hx].
    case_sepcon H.
    have Hn2 : 0 < [n2 ]_ s.
      inversion H_h2; last by done.
      subst.
      rewrite /= in H1.
      by rewrite H1 /= in Hx.
    move/ls_cons_inv : H_h2.
    move/(_ Hn2).
    case=> h21 [h22 [v' [Hdisj' [Hunion' [[h211 [h212 [Hdisj'' [Hunion'' [Hh211 [p [Hp Hh212]]]]]]] Hh22]]]]].
    exists ([ v' ]e_ s); split.
    * rewrite h1_U_h2 Hunion' Hunion''.
      apply heap.get_union_R; first by map_tac_m.Disj.
      apply heap.get_union_L; first by map_tac_m.Disj.
      apply heap.get_union_R; first by map_tac_m.Disj.
      by rewrite Hh212 Hp heap.get_sing.
    * rewrite /formula_interp' /=.
      exists v.
      Compose_sepcon (h1 \U h21) h22.
      have H121 : ls (var_e n1 \+ cst_e 1)%seplog_expr (cst_e v) v' s (h1 \U h21).
        apply ls_app_cons with (var_e x) => //.
        by map_tac_m.Disj.
        rewrite Hunion''; apply con_cons => //.
        by exists p.
      eapply ls_equiv; first by apply H121.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
    * eapply ls_equiv; first by apply Hh22.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
  + apply abst_inst_assign with
      (ls_Gamma_1 x (fvar_e n1) (fvar_e n2 \- cst_e' 1)).
    by apply varset_m.mem_1, varset_m.add_1.
    * apply hoare_def.
      apply hoare_assign' => s h.
      rewrite /formula_interp [formula_size _]/= [formula_interp' _ _]/=.
      case => v [h1 [h2 [Hdisj [Hunion [H1 H2]]]]].
      rewrite /wp_assign /=.
      exists v.
      rewrite Hunion; apply con_cons => //.
      eapply ls_equiv; first by apply H1.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
      eapply ls_equiv; first by apply H2.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
        by rewrite /=; repeat Store_upd => //.
    * apply abst_inst_assign with (ls_Gamma_1 x (fvar_e n1) (fvar_e n2)).
      by apply varset_m.mem_1, varset_m.add_2, varset_m.singleton_2.
      - apply hoare_def.
        apply hoare_assign' => s h.
        rewrite /formula_interp [formula_size _]/= [formula_interp' _ _]/=.
        case=> v [h1 [h2 [Hdisj [Hunion [H1 H2]]]]].
        rewrite /wp_assign /=.
        exists v.
        rewrite Hunion; apply con_cons => //.
        eapply ls_equiv; first by apply H1.
          by rewrite /=; repeat Store_upd => //.
          by rewrite /=; repeat Store_upd => //.
          by rewrite /=; repeat Store_upd => //.
        eapply ls_equiv; first by apply H2.
          by repeat (rewrite /=; Store_upd).
          by repeat (rewrite /=; Store_upd).
          by repeat (rewrite /=; Store_upd).
      - by apply abst_goto.
- destruct i; last by rewrite /= in Hi.
  by apply abst_halt.
Qed.

(** second example of section 5.3 *)

Fixpoint lift_expr' e :=
  match e with
    | fvar_e x => fvar_e x
    | bvar_e i => bvar_e i.+1
    | cst_e' v => cst_e' v
    | bop_e' b e1' e2' => bop_e' b (lift_expr' e1') (lift_expr' e2')
    | uop_e' a e1 => uop_e' a (lift_expr' e1)
  end.

Fixpoint lift_expr_b' eb :=
  match eb with
    | true_b' => true_b'
    | bop_b' b e1' e2' => bop_b' b (lift_expr' e1') (lift_expr' e2')
    | neg_b' e1' => neg_b' (lift_expr_b' e1')
    | bop_b2' b e1' e2' => bop_b2' b (lift_expr_b' e1') (lift_expr_b' e2')
  end.

Definition lift_spatial f :=
  match f with
    | e1 |~> e2 => (lift_expr' e1) |~> (lift_expr' e2)
    | e1 |~>_ => (lift_expr' e1) |~>_
    | spa_emp => spa_emp
    | spa_ls n e1 e2 => spa_ls (lift_expr' n) (lift_expr' e1) (lift_expr' e2)
    | spa_tree e1 e2 => spa_tree (lift_expr' e1) (lift_expr' e2)
  end.

Lemma for_exists_swap_b : forall (e : expr_b') (g : formula),
  closed_b e ->
  formula_interp (for_bool e ** for_exists g) ===>
  formula_interp (for_exists (for_bool e ** g)).
Proof.
elim=> //=.
- move=> g _.
  rewrite {1}/formula_interp /=.
  move=> s h H.
  case_sepcon H.
  case: H_h2 => v H_h2.
  exists v.
  Compose_sepcon h1 h2.
  simpl.
  by destruct (formula_size g).
  simpl.
  done.
- move=> b e1 e2 g Hno_bvar.
  rewrite {1}/formula_interp /=.
  move=> s h H.
  case_sepcon H.
  case: H_h2 => v H_h2.
  exists v.
  Compose_sepcon h1 h2.
  simpl.
  case/closed_b_bop_b' : Hno_bvar => Hno_bvar1 Hno_bvar2.
  rewrite subst_e'_free // subst_e'_free //.
  by destruct (formula_size g).
  done.
- admit.
- admit.
Admitted.

Lemma for_exists_swap_b2 : forall f g,
  closed_b f ->
  formula_interp (for_bool f ** for_exists (for_exists g)) ===>
  formula_interp (for_exists (for_exists (for_bool f ** g))).
Proof.
move=> f g Hf.
eapply hoare_prop_m.entails_trans.
by apply for_exists_swap_b.
Admitted.

Definition tree_prg (r : var.v) :=
  cont_branch
  ((var_e r \!= NIL) :: (var_e r \= NIL) :: nil)%seplog_expr
  (cont_branch 
    (true_b :: true_b :: nil)
    ((r <-* r -.> left ; cont_goto 1) :: (r <-* r -.> right; cont_goto 1) :: nil)
    :: cont_halt :: nil)%seplog_expr.

Definition instrumented_tree_prg (r n1 n2 n : var.v) :=
  cont_branch
  ((var_e r \!= NIL) :: (var_e r \= NIL) :: nil)%seplog_expr
  ((n1 <-?; n2 <-?;
    assume (var_e n1 \< var_e n \&& var_e n2 \< var_e n)%seplog_expr ;; cont_branch 
    (true_b :: true_b :: nil)
    ((r <-* r -.> left ; n <- var_e n1; cont_goto 1) :: (r <-* r -.> right; n <- var_e n2; cont_goto 1) :: nil))
    :: cont_halt :: nil)%seplog_expr.

Definition tree_Gamma_1 (r : expr') (n : expr') : formula := 
  (for_bool true_b' ** (for_spa (spa_tree n r))).

Definition Q' (r n:var.v) (n1 n2 : (*nat*)expr') := 
  for_bool (fvar_e n >> cst_e' 0) //\\
  for_bool ((*bvar_e*) n1 << fvar_e n) //\\
  for_bool ((*bvar_e*) n2 << fvar_e n) //\\
  for_exists (for_exists
    ((for_spa (fvar_e r |~>_) ** 
      (for_spa (r -.> left |~> bvar_e 1 (*lc*))) **
      (for_spa (r -.> right |~> bvar_e 0 (*rc*))) **
      (for_spa (spa_tree ((*bvar_e*) (lift_expr' (lift_expr' n1))) (bvar_e 1 (*lc*)))) **
      (for_spa (spa_tree ((*bvar_e*) (lift_expr' (lift_expr' n2))) (bvar_e 0 (*rc*))))))).

Lemma coucou : forall x,
  formula_interp (for_bool (fvar_e x <>e cst_e' 0)) ===>
  formula_interp (for_exists (for_bool (bvar_e 0 <>e cst_e' 0))).
Proof.
move=> x s h /=.
rewrite /formula_interp /=.
move/negP => H.
exists ([x ]_ s).
by apply/negP.
Abort.

Definition Q (r n : var.v) := 
  for_exists (for_exists (Q' r n (bvar_e 1) (*n1*) (bvar_e 0)(*n2*))).

Lemma example_53b : forall (r n1 n2 n : var.v), 
  uniq (r :: n1 :: n2 :: n :: nil) ->
  (1%nat, tree_Gamma_1 (fvar_e r) (fvar_e n)) :: nil |-  tree_Gamma_1 (fvar_e r) (fvar_e n) -| 
    instrumented_tree_prg r n1 n2 n <| varset_m.add n (varset_m.add n1 (varset_m.singleton n2)) |> tree_prg r.
Proof.
move=> r n1 n2 n Hset.
rewrite /tree_prg /instrumented_tree_prg.
(* we apply the BRANCH rule, what gives two subgoals *)
apply abst_branch => // i Hi.
destruct i.
- move=> {Hi} /=.
  rewrite {2}/tree_Gamma_1.
  (* we apply the STRENGTHENING rule several times *)
  apply abst_stren with (for_bool true_b' ** 
    (for_spa (spa_tree (fvar_e n) (fvar_e r)) //\\ for_bool (fvar_e r <>e cst_e' 0))).
    rewrite /formula_interp /=.
    by apply con_and_pure.
  apply abst_stren with (for_bool true_b' ** Q r n).
    rewrite /Q /Q'.
    apply hoare_prop_m.entails_trans with (formula_interp (for_bool true_b' ** 
      (for_exists (*n1*)
        (for_exists (*n2*)
          (for_exists (*lc*)
            (for_exists (*rc*)
              (for_and
                (for_bool (cst_e' 0 << fvar_e n &e bvar_e 3(*n1*) << fvar_e n &e bvar_e 2(*n2*) << fvar_e n))
                ((for_spa (fvar_e r |~>_)) ** 
                  (for_spa (r -.> left |~> bvar_e 1(*lc*))) ** 
                  (for_spa (r -.> right |~> bvar_e 0(*rc*))) ** 
                  (for_spa (spa_tree (bvar_e 3) (bvar_e 1(*lc*)))) ** 
                  (for_spa (spa_tree (bvar_e 2) (bvar_e 0(*rc*)))))))))))).
    (* one unrolling of the formula defining a tree *)
    rewrite /formula_interp /=.
    apply monotony' => //.
    move: (tree_unroll n r) => Hunroll.
    rewrite /formula_interp /= in Hunroll.
    move=> s h H.
    case: H => H1 H2.
    move: {Hunroll}(Hunroll _ _ H1).
    case.
    case/andP.
    by rewrite (negbTE H2).
    case => H.
    case => v Hv; exists v.
    case: Hv => v0 Hv0; exists v0.
    case: Hv0 => Hv01 Hv02.
    case: Hv02 => v1 Hv1; exists v1.
    case: Hv1 => v2 Hv2; exists v2.
    split; last by done.
    rewrite -andbA Hv01 andbC /=.
    apply/ZgeP.
    move/ZgtP in H; omega.
    (* move all for_exists upfront *)
    admit. 
  rewrite /Q.
  apply abst_stren with (for_exists (for_exists (for_bool true_b' ** (Q' r n (bvar_e 1) (bvar_e 0))))).
    apply for_exists_swap_b2.
    done.
  (* we apply the INST-EXISTS rule twice *)
  apply abst_inst_exists with (x' := n1) (L := r :: n :: nil).
    by Uniq_not_In.
    by apply varset_m.mem_1, varset_m.add_2, varset_m.add_1.
  apply abst_inst_exists with (x' := n2) (L := n1 :: r :: n :: nil).
    by Uniq_not_In.
    by apply varset_m.mem_1, varset_m.add_2, varset_m.add_2, varset_m.singleton_2.
  (* we apply the INST-ASSUME rule *)
  apply abst_inst_assume.
    rewrite /Q' /=.
    admit.
  (* we apply the BRANCH rule, what gives two subgoals *)
  apply abst_branch => // i Hi.
  destruct i.
  - (* we apply the COMMAND rule *)
    apply abst_command with (for_bool true_b' ** for_spa (spa_tree (fvar_e n1) (fvar_e r))).
    rewrite /=.
    apply hoare_def.
    apply hoare_lookup'.
    rewrite /wp_lookup /formula_interp /= => s h H.
    case: H => H _.
    case_sepcon H.
    clear H_h1.
    case: H_h2 => [H1 [H2 [H3 [lc [rc Hh2]]]]].
    rewrite -!conAE in Hh2.
    case_sepcon Hh2.
    case_sepcon Hh2_h21.
    case_sepcon Hh2_h21_h211.
    case_sepcon Hh2_h21_h211_h2111.
    exists lc.
    split.
    rewrite h1_U_h2 h21_U_h22 h211_U_h212 h2111_U_h2112 h21111_U_h21112.
    apply heap.get_union_R; first by map_tac_m.Disj.
    apply heap.get_union_L; first by map_tac_m.Disj.
    apply heap.get_union_L; first by map_tac_m.Disj.
    apply heap.get_union_L; first by map_tac_m.Disj.
    apply heap.get_union_R; first by map_tac_m.Disj.
    case: Hh2_h21_h211_h2111_h21112 => X Y.
    case: X => p [X1 X2].
    rewrite X2.
    rewrite /= in X1.
    by rewrite X1 heap.get_sing.
    Compose_sepcon (h1 \U (((h21111 \U h2112) \U h21112) \U h22)) h212; first by done.
    eapply tree_equiv; first by apply Hh2_h21_h212.
    by rewrite /=; Store_upd.
    by rewrite /=; Store_upd.
    (* we apply the INST-ASSIGN rule *)
    apply abst_inst_assign with (for_bool true_b' ** for_spa (spa_tree (fvar_e n) (fvar_e r))).
      by apply varset_m.mem_1, varset_m.add_1.
    apply hoare_def.
    apply hoare_assign'.
    rewrite /wp_assign /formula_interp /=.
    apply monotony'; first by done.
    rewrite /while.entails => s h H.
    eapply tree_equiv; eauto.
    by rewrite /=; Store_upd.
    by rewrite /=; Store_upd.
    (* we apply the GOTO rule *)
    apply abst_goto => /=.
    by rewrite /tree_Gamma_1.
  - destruct i; last by rewrite /= in Hi.
    move=> {Hi}.
    (* we apply the COMMAND rule *)
    apply abst_command with (for_bool true_b' ** (for_spa (spa_tree (fvar_e n2) (fvar_e r)))).
    admit.
    (* we apply the INST-ASSIGN rule *)
    apply abst_inst_assign with (for_bool true_b' ** for_spa (spa_tree (fvar_e n) (fvar_e r))).
      by apply varset_m.mem_1, varset_m.add_1.
    admit.
    (* we apply the GOTO rule *)
    apply abst_goto => /=.
    by rewrite /tree_Gamma_1.
- destruct i; last by rewrite /= in Hi.
  by apply abst_halt.
Admitted.
